# -*- coding: utf-8 -*-

"""Implementation of TagManager."""

import asyncio
import datetime
from typing import (
    Any,
    Awaitable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from nisystemlink.clients import core, tag as tbase
from nisystemlink.clients.core._internal._http_client import HttpClient, HttpResponse
from nisystemlink.clients.core._internal._timestamp_utilities import TimestampUtilities
from nisystemlink.clients.tag._core._manual_reset_timer import ManualResetTimer
from nisystemlink.clients.tag._core._serialized_tag_with_aggregates import (
    SerializedTagWithAggregates,
)
from nisystemlink.clients.tag._core._system_time_stamper import SystemTimeStamper
from nisystemlink.clients.tag._http._http_async_tag_query_result_collection import (
    HttpAsyncTagQueryResultCollection,
)
from nisystemlink.clients.tag._http._http_buffered_tag_writer import (
    HttpBufferedTagWriter,
)
from nisystemlink.clients.tag._http._http_tag_query_result_collection import (
    HttpTagQueryResultCollection,
)
from nisystemlink.clients.tag._http._http_tag_selection import HttpTagSelection
from nisystemlink.clients.tag._http._temporary_tag_selection import (
    TemporaryTagSelection,
)
from typing_extensions import final


@final
class TagManager(tbase.ITagReader):
    """Represents common ways to create, read, and query SystemLink tags for a specific server connection."""

    def __init_subclass__(cls) -> None:
        raise TypeError("type 'TagManager' is not an acceptable base type")

    def __init__(self, configuration: Optional[core.HttpConfiguration] = None) -> None:
        """Initialize an instance.

        Args:
            configuration: Defines the web server to connect to and information about
                how to connect.

        Raises:
            ApiException: if the current system cannot communicate with a SystemLink
                Server, or if the configuration provided by SystemLink Client cannot be
                found.
        """
        if configuration is None:
            configuration = core.HttpConfigurationManager.get_configuration()

        self._http_client = HttpClient(configuration)
        self._api = self._http_client.at_uri("/nitag/v2")

    def create_selection(self, tags: List[tbase.TagData]) -> tbase.TagSelection:
        """Create an :class:`TagSelection` that initially contains the given ``tags``
        without retrieving any additional data from the server.

        Args:
            tags: The tags to include in the selection.

        Returns:
            The created selection.

        Raises:
            ValueError: if any of the given ``tags`` is None or has an invalid path.
            ValueError: if ``tags`` is None.
        """
        return HttpTagSelection(self._http_client, tags)

    def open_selection(self, paths: List[str]) -> tbase.TagSelection:
        """Query the server for the metadata for the given tag ``paths`` and return the
        results in a :class:`TagSelection`.

        Args:
            paths: The paths of the tags to include in the selection. May include
                glob-style wildcards.

        Returns:
            The created selection with :attr:`TagSelection.metadata` containing the
            metadata.

        Raises:
            ValueError: if any of the given ``paths`` is None or invalid.
            ValueError: if ``paths`` contains duplicate paths.
            ValueError: if ``paths`` is None.
            ApiException: if the API call fails.
        """
        return HttpTagSelection.open(self._http_client, paths)

    def open_selection_async(self, paths: List[str]) -> Awaitable[tbase.TagSelection]:
        """Asynchronously query the server for the metadata for the given tag ``paths``
        and return the results in a :class:`TagSelection`.

        Args:
            paths: The paths of the tags to include in the selection. May include
                glob-style wildcards.

        Returns:
            A task representing the asynchronous operation. On success, contains the
            created selection with :attr:`TagSelection.metadata` containing the
            metadata.

        Raises:
            ValueError: if any of the given ``paths`` is None or invalid.
            ValueError: if ``paths`` contains duplicate paths.
            ValueError: if ``paths`` is None.
            ApiException: if the API call fails.
        """
        return HttpTagSelection.open_async(self._http_client, paths)

    def open(
        self,
        path: str,
        data_type: Optional[tbase.DataType] = None,
        *,
        create: Optional[bool] = None
    ) -> tbase.TagData:
        """Query the server for the metadata of a tag, optionally creating it if it doesn't already exist.

        If ``data_type`` is provided, ``create`` defaults to True. If ``data_type`` is
        not provided, ``create`` cannot be set to True.

        The call fails if the tag already exists as a different data type than specified
        or if it doesn't exist and ``create`` is False.

        Args:
            path: The path of the tag to open.
            data_type: The expected data type of the tag.
            create: True to create the tag if it doesn't already exist, False to fail if
                it doesn't exist.

        Returns:
            Information about the tag.

        Raises:
            ValueError: if ``path`` is None or empty.
            ValueError: if ``data_type`` is invalid.
            ValueError: if ``create`` is True, but ``data_type`` is None.
            ApiException: if the API call fails.
        """
        if create is None:
            create = data_type is not None
        elif create is True:
            if data_type is None:
                raise ValueError("Cannot create if data_type is not specified")

        if data_type == tbase.DataType.UNKNOWN:
            raise ValueError("Must specify a valid data type")

        tag = None  # type: Optional[Dict[str, Any]]
        try:
            tag, _ = self._api.get(
                "/tags/{path}", params={"path": tbase.TagPathUtilities.validate(path)}
            )
        except core.ApiException as ex:
            error_name = None if ex.error is None else ex.error.name
            if create and (error_name or "").startswith("Tag.NoSuchTag"):
                pass  # continue on and create the tag
            else:
                raise

        if tag is not None:
            if data_type is not None and tag["type"] != data_type.api_name:
                raise core.ApiException("Tag exists with a conflicting data type")

            return tbase.TagData.from_json_dict(tag)
        else:
            if data_type is None:
                raise ValueError("data_type cannot be None when create is True")

            # Tag didn't already exist, so try to create it.
            self._api.post("/tags", data={"type": data_type.api_name, "path": path})
            return tbase.TagData(path, data_type)

    async def open_async(
        self,
        path: str,
        data_type: Optional[tbase.DataType] = None,
        *,
        create: Optional[bool] = None
    ) -> tbase.TagData:
        """Asynchronously query the server for the metadata of a tag, optionally
        creating it if it doesn't already exist.

        The call fails if the tag already exists as a different data type than specified
        or if it doesn't exist and ``create`` is False.

        Args:
            path: The path of the tag to open.
            data_type: The expected data type of the tag.
            create: True to create the tag if it doesn't already exist, False to fail if
                it doesn't exist.

        Returns:
            A task representing the asynchronous operation. On success, contains
            information about the tag.

        Raises:
            ValueError: if ``path`` is None or empty.
            ValueError: if ``data_type`` is invalid.
            ValueError: if ``create`` is True, but ``data_type`` is None.
            ApiException: if the API call fails.
        """
        if create is None:
            create = data_type is not None
        elif create is True:
            if data_type is None:
                raise ValueError("Cannot create if data_type is not specified")

        if data_type == tbase.DataType.UNKNOWN:
            raise ValueError("Must specify a valid data type")

        tag = None  # type: Optional[Dict[str, Any]]
        try:
            tag, _ = await self._api.as_async.get(
                "/tags/{path}", params={"path": tbase.TagPathUtilities.validate(path)}
            )
        except core.ApiException as ex:
            error_name = None if ex.error is None else ex.error.name
            if create and (error_name or "").startswith("Tag.NoSuchTag"):
                pass  # continue on and create the tag
            else:
                raise

        if tag is not None:
            if data_type is not None and tag["type"] != data_type.api_name:
                raise core.ApiException("Tag exists with a conflicting data type")

            return tbase.TagData.from_json_dict(tag)
        else:
            if data_type is None:
                raise ValueError("data_type cannot be None when create is True")

            # Tag didn't already exist, so try to create it.
            await self._api.as_async.post(
                "/tags", data={"type": data_type.api_name, "path": path}
            )
            return tbase.TagData(path, data_type)

    def refresh(self, tags: List[tbase.TagData]) -> None:
        """Populate the given ``tags`` with the latest metadata from the server.

        Only the :attr:`TagData.path` needs to be initialized. Tags that don't exist on
        the server will have their :attr:`TagData.data_type` set to
        :attr:`DataType.UNKNOWN`.

        Args:
            tags: The tags to refresh.

        Raises:
            ValueError: if any ``tags`` are None or have invalid paths.
            ValueError: if ``tags`` is None.
            ApiException: if the API call fails.
        """
        paths = self._prepare_refresh(tags)
        response, http_response = self._api.get(
            "/tags", params={"path": paths, "take": str(len(tags))}
        )
        self._handle_refresh(tags, response, http_response)

    async def refresh_async(self, tags: List[tbase.TagData]) -> None:
        """Asynchronously populate the given ``tags`` with the latest metadata from the server.

        Only the :attr:`TagData.path` needs to be initialized. Tags that don't exist on
        the server will have their :attr:`TagData.data_type` set to
        :attr:`DataType.UNKNOWN`.

        Args:
            tags: The tags to refresh.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ValueError: if any ``tags`` are None or have invalid paths.
            ValueError: if ``tags`` is None.
            ApiException: if the API call fails.
        """
        paths = self._prepare_refresh(tags)
        response, http_response = await self._api.as_async.get(
            "/tags", params={"path": paths, "take": str(len(tags))}
        )
        self._handle_refresh(tags, response, http_response)

    def _prepare_refresh(self, tags: List[tbase.TagData]) -> str:
        if tags is None:
            raise ValueError("tags cannot be None")
        elif any(t is None for t in tags):
            raise ValueError("Tags cannot contain None")

        return ",".join(t.validate_path() for t in tags)

    def _handle_refresh(
        self,
        tags: List[tbase.TagData],
        refresh_result: Dict[str, Any],
        http_response: HttpResponse,
    ) -> None:
        if refresh_result is None or refresh_result.get("tags") is None:
            raise self.invalid_response(http_response)

        read_tags = {t["path"]: t for t in refresh_result["tags"]}

        for data in tags:
            tag = read_tags.get(data.path)
            if tag is not None:
                data.clear_retention()
                data.data_type = tbase.DataType.from_api_name(
                    tag.get("type") or "UNKNOWN"
                )
                data.replace_keywords(tag.get("keywords") or [])
                data.replace_properties(tag.get("properties") or {})
                data.collect_aggregates = tag.get("collectAggregates") or False
            else:
                data.data_type = tbase.DataType.UNKNOWN

    def query(
        self,
        paths: Optional[Sequence[str]] = None,
        keywords: Optional[Iterable[str]] = None,
        properties: Optional[Dict[str, str]] = None,
        *,
        skip: int = 0,
        take: Optional[int] = None
    ) -> tbase.TagQueryResultCollection:
        """Query the server for available tags matching the given criteria.

        Args:
            paths: List of tag paths to include in the result. May include glob-style
                wildcards.
            keywords: List of keywords that tags must have, or None.
            properties: Mapping of properties and their values that tags must have, or
                None.
            skip: The number of tags to initially skip in the results.
            take: The number of tags to include in each page of results.

        Returns:
            A :class:`TagQueryResultCollection` containing the first page of results.
            Enumerating the collection will retrieve additional pages according to the
            ``take`` parameter.

        Raises:
            ValueError: if ``skip`` or ``take`` is negative.
            ValueError: if ``paths`` is an empty list.
            ValueError: if any of ``paths`` are None.
            ApiException: if the API call fails.
        """
        path_str, keyword_str, prop_str = self._prepare_query(
            paths, keywords, properties, skip, take
        )
        params = {
            "path": path_str,
            "keywords": keyword_str,
            "properties": prop_str,
            "skip": str(skip) if skip is not None else None,
            "take": str(take) if take is not None else None,
        }
        for k, v in list(params.items()):
            if v is None:
                del params[k]
        first_page, http_response = self._api.get("/tags", params=params)
        return HttpTagQueryResultCollection(
            self._http_client,
            path_str,
            keyword_str,
            prop_str,
            skip,
            take,
            first_page,
            http_response,
        )

    async def query_async(
        self,
        paths: Optional[Sequence[str]] = None,
        keywords: Optional[Iterable[str]] = None,
        properties: Optional[Dict[str, str]] = None,
        *,
        skip: int = 0,
        take: Optional[int] = None
    ) -> tbase.AsyncTagQueryResultCollection:
        """Asynchronously query the server for available tags matching the given criteria.

        Args:
            paths: List of tag paths to include in the result. May include glob-style
                wildcards.
            keywords: List of keywords that tags must have, or None.
            properties: Mapping of properties and their values that tags must have, or
                None.
            skip: The number of tags to initially skip in the results.
            take: The number of tags to include in each page of results.

        Returns:
            A task representing the asynchronous operation. On success, contains a
            :class:`TagQueryResultCollection` containing the first page of results.
            Enumerating the collection will retrieve additional pages according to the
            ``take`` parameter.

        Raises:
            ValueError: if ``skip`` is negative.
            ValueError: if ``take`` is negative.
            ApiException: if the API call fails.
        """
        path_str, keyword_str, prop_str = self._prepare_query(
            paths, keywords, properties, skip, take
        )
        params = {
            "path": path_str,
            "keywords": keyword_str,
            "properties": prop_str,
            "skip": str(skip) if skip is not None else None,
            "take": str(take) if take is not None else None,
        }
        for k, v in list(params.items()):
            if v is None:
                del params[k]
        first_page, http_response = await self._api.as_async.get("/tags", params=params)
        return HttpAsyncTagQueryResultCollection(
            self._http_client,
            path_str,
            keyword_str,
            prop_str,
            skip,
            take,
            first_page,
            http_response,
        )

    def _prepare_query(
        self,
        paths: Optional[Sequence[str]],
        keywords: Optional[Iterable[str]],
        properties: Optional[Dict[str, str]],
        skip: Optional[int],
        take: Optional[int] = None,
    ) -> Tuple[Optional[str], Optional[str], Optional[str]]:
        if paths is not None and len(paths) == 0:
            raise ValueError("paths cannot be empty an empty list")
        if paths is not None and any(p is None for p in paths):
            raise ValueError("paths cannot contain None")
        if skip is not None and skip < 0:
            raise ValueError("skip cannot be negative")
        if take is not None and take < 0:
            raise ValueError("take cannot be negative")

        path_str = None
        keyword_str = None
        prop_str = None

        if paths is not None:
            path_str = ",".join(tbase.TagPathUtilities.validate_query(p) for p in paths)
        if keywords:
            keyword_str = ",".join(keywords)
        if properties:
            prop_str = ",".join("{}={}".format(k, v) for k, v in properties.items())

        return path_str, keyword_str, prop_str

    def update(
        self, updates: Union[Sequence[tbase.TagData], Sequence[tbase.TagDataUpdate]]
    ) -> None:
        """Update the metadata of one or more tags on the server, creating tags that don't exist.

        If ``updates`` contains :class:`TagData` objects, existing metadata will be
        replaced. If ``updates`` contains :class:`TagDataUpdate` objects instead, tags
        that already exist will have their existing keywords, properties, and settings
        merged with those specified in the corresponding :class:`TagDataUpdate`.

        The call fails if any of the tags already exist as a different data type.

        Args:
            updates: The tags to update (if :class:`TagData` objects are given), or the
                tag metadata updates to send (if :class:`TagDataUpdate` objects are
                given).

        Raises:
            ValueError: if ``updates`` is None or empty.
            ValueError: if ``updates`` contains any invalid tags.
            ValueError: if ``updates`` contains both ``TagData`` objects and
                ``TagDataUpdate`` objects.
            ApiException: if the API call fails.
        """
        tag_models, merge = self._prepare_update(updates)
        partial_success, _ = self._api.post(
            "/update-tags", data={"tags": tag_models, "merge": merge}
        )

        if partial_success is not None:
            err_dict = partial_success.get("error")
            if err_dict is None and "code" in partial_success:
                # SystemLink Cloud has a bug in which it returns the error directly
                # instead of under the "error" key
                err_dict = partial_success
            err_obj = core.ApiError.parse_obj(err_dict) if err_dict else None
            if err_dict is None:
                assert False, partial_success
            raise core.ApiException(error=err_obj)

    async def update_async(
        self, updates: Union[Sequence[tbase.TagData], Sequence[tbase.TagDataUpdate]]
    ) -> None:
        """Asynchronously update the metadata of one or more tags on the server, creating tags that don't exist.

        If ``updates`` contains :class:`TagData` objects, existing metadata will be
        replaced. If ``updates`` contains :class:`TagDataUpdate` objects instead, tags
        that already exist will have their existing keywords, properties, and settings
        merged with those specified in the corresponding :class:`TagDataUpdate`.

        Args:
            updates: The tags to update (if :class:`TagData` objects are given), or the
                tag metadata updates to send (if :class:`TagDataUpdate` objects are
                given).

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ValueError: if ``updates`` is None or empty.
            ValueError: if ``updates`` contains any invalid tags.
            ValueError: if ``updates`` contains both ``TagData`` objects and
                ``TagDataUpdate`` objects.
            ApiException: if the API call fails.
        """
        tag_models, merge = self._prepare_update(updates)
        partial_success, _ = await self._api.as_async.post(
            "/update-tags", data={"tags": tag_models, "merge": merge}
        )

        if partial_success is not None:
            err_dict = partial_success.get("error")
            if err_dict is None and "code" in partial_success:
                # SystemLink Cloud has a bug in which it returns the error directly
                # instead of under the "error" key
                err_dict = partial_success
            err_obj = core.ApiError.parse_obj(err_dict) if err_dict else None
            if err_dict is None:
                assert False, partial_success
            raise core.ApiException(error=err_obj)

    def _prepare_update(
        self, updates: Union[Sequence[tbase.TagData], Sequence[tbase.TagDataUpdate]]
    ) -> Tuple[List[Dict[str, Any]], bool]:
        if updates is None:
            raise ValueError("updates cannot be None")
        if not updates:
            raise ValueError("updates cannot be empty")
        if not all(isinstance(u, type(updates[0])) for u in updates):
            raise ValueError(
                "updates must contain only TagData objects or TagDataUpdate objects, not both"
            )

        merge = isinstance(updates[0], tbase.TagDataUpdate)

        return [u.to_json_dict() for u in updates], merge

    def delete(self, tags: Iterable[Union[tbase.TagData, str]]) -> None:
        """Delete one or more tags from the server.

        Args:
            tags: The tags (or tag paths) to delete.

        Raises:
            ValueError: if ``tags`` is None.
            ValueError: if ``tags`` contains any invalid tags.
            ApiException: if the API call fails.
        """
        if tags is None:
            raise ValueError("tags cannot be None")
        if any(t is None for t in tags):
            raise ValueError("tags cannot contain None")

        validated_paths = [
            (
                tbase.TagPathUtilities.validate(t)
                if isinstance(t, str)
                else t.validate_path()
            )
            for t in tags
        ]

        self._perform_delete(validated_paths)

    def delete_async(
        self, tags: Iterable[Union[tbase.TagData, str]] = None
    ) -> Awaitable[None]:
        """Asynchronously delete one or more tags from the server.

        Args:
            tags: The tags (or tag paths) to delete.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ValueError: if ``tags`` is None.
            ValueError: if ``tags`` contains any invalid tags.
            ApiException: if the API call fails.
        """
        try:
            if tags is None:
                raise ValueError("tags cannot be None")
            if any(t is None for t in tags):
                raise ValueError("tags cannot contain None")

            validated_paths = [
                (
                    tbase.TagPathUtilities.validate(t)
                    if isinstance(t, str)
                    else t.validate_path()
                )
                for t in tags
            ]
        except Exception as ex:
            f = asyncio.get_event_loop().create_future()
            f.set_exception(ex)
            return f

        return self._perform_delete_async(validated_paths)

    def _perform_delete(self, paths: List[str]) -> None:
        if len(paths) < 4:
            # Few enough to make multiple, single deletes rather than creating a selection.
            exceptions = []

            for path in paths:
                try:
                    self._api.delete(
                        "/tags/{path}",
                        params={"path": tbase.TagPathUtilities.validate(path)},
                    )
                except core.ApiException as ex:
                    exceptions.append(ex)

            if exceptions:
                raise exceptions[0] from None
        else:
            with TemporaryTagSelection.create(self._http_client, paths) as selection:
                self._api.delete("/selections/{id}/tags", params={"id": selection.id})

    async def _perform_delete_async(self, paths: List[str]) -> None:
        if len(paths) < 4:
            # Few enough to make multiple, single deletes rather than creating a selection.
            await asyncio.gather(
                *[
                    self._api.as_async.delete(
                        "/tags/{path}",
                        params={"path": tbase.TagPathUtilities.validate(p)},
                    )
                    for p in paths
                ]
            )
        else:
            async with await TemporaryTagSelection.create_async(
                self._http_client, paths
            ) as selection:
                await self._api.as_async.delete(
                    "/selections/{id}/tags", params={"id": selection.id}
                )

    def create_writer(
        self,
        *,
        buffer_size: Optional[int] = None,
        max_buffer_time: Optional[datetime.timedelta] = None
    ) -> tbase.BufferedTagWriter:
        """Create a tag writer that buffers tag values until
        :meth:`~BufferedTagWriter.send_buffered_writes()` is called on the returned
        object, ``buffer_size`` writes have been buffered, or ``max_buffer_time`` time
        has past since buffering a value, at which point the writes will be sent
        automatically.

        Args:
            buffer_size: The maximum number of tag writes to buffer before automatically
                sending them to the server.
            max_buffer_time: The amount of time before writes are sent.

        Returns:
            The created writer. Close the writer to free resources.

        Raises:
            ValueError: if ``buffer_size`` and ``max_buffer_time`` are both None.
            ValueError: if ``buffer_size`` is less than one.
        """
        if buffer_size is None and max_buffer_time is None:
            raise ValueError("must provide either buffer_size or max_buffer_time")

        if buffer_size is not None:
            if buffer_size < 1:
                raise ValueError("buffer_size cannot be 0 or negative")
        else:
            buffer_size = 0

        if max_buffer_time is not None:
            if max_buffer_time.total_seconds() < 0.001:
                raise ValueError("max_buffer_time must be at least 1 millisecond")
            timer = ManualResetTimer(max_buffer_time)
        else:
            timer = ManualResetTimer.null_timer

        return HttpBufferedTagWriter(
            self._http_client, SystemTimeStamper(), buffer_size, timer
        )

    def _read(
        self, path: str, include_timestamp: bool, include_aggregates: bool
    ) -> Optional[SerializedTagWithAggregates]:
        """Retrieve the current value of the tag with the given ``path`` from the server.

        Optionally retrieves the aggregate values as well. The tag must exist. Clients
        do not typically call this method directly. Use a :class:`.TagValueReader`
        instead.

        Args:
            path: The path of the tag to read.
            include_timestamp: True to include the timestamp associated with the value
                in the result.
            include_aggregates: True to include the tag's aggregate values in the result
                if the tag is set to :attr:`TagData.collect_aggregates`.

        Returns:
            The value serialized as a string, along with timestamp and/or aggregates, if
            requested, or None if the tag exists but doesn't have a value.

        Raises:
            ValueError: if ``path`` is empty or invalid.
            ValueError: if ``path`` is None.
            ApiException: if the API call fails.
        """
        path = tbase.TagPathUtilities.validate(path)

        if include_aggregates:
            response, http_response = self._api.get(
                "/tags/{path}/values", params={"path": path}
            )
            return self._handle_read(
                path, response, http_response, include_timestamp, True
            )
        elif include_timestamp:
            response2, http_response = self._api.get(
                "/tags/{path}/values/current", params={"path": path}
            )
            return self._handle_read(
                path, response2, http_response, include_timestamp, False
            )
        else:
            response3, http_response = self._api.get(
                "/tags/{path}/values/current/value", params={"path": path}
            )
            return self._handle_read(path, response3, http_response)

    async def _read_async(
        self, path: str, include_timestamp: bool, include_aggregates: bool
    ) -> Optional[SerializedTagWithAggregates]:
        """Asynchronously retrieve the current value of the tag with the given ``path`` from the server.

        Optionally retrieves the aggregate values as well. The tag must exist. Clients
        do not typically call this method directly. Use a :class:`.TagValueReader`
        instead.

        Args:
            path: The path of the tag to read.
            include_timestamp: True to include the timestamp associated with the value
                in the result.
            include_aggregates: True to include the tag's aggregate values in the result
                if the tag is set to :attr:`TagData.collect_aggregates`.

        Returns:
            The value serialized as a string, or None if the tag exists but doesn't have
            a value.

        Raises:
            ValueError: if ``path`` is empty or invalid.
            ValueError: if ``path`` is None.
            ApiException: if the API call fails.
        """
        path = tbase.TagPathUtilities.validate(path)

        if include_aggregates:
            response, http_response = await self._api.as_async.get(
                "/tags/{path}/values", params={"path": path}
            )
            return self._handle_read(
                path, response, http_response, include_timestamp, True
            )
        elif include_timestamp:
            response2, http_response = await self._api.as_async.get(
                "/tags/{path}/values/current", params={"path": path}
            )
            return self._handle_read(
                path, response2, http_response, include_timestamp, False
            )
        else:
            response3, http_response = await self._api.as_async.get(
                "/tags/{path}/values/current/value", params={"path": path}
            )
            return self._handle_read(path, response3, http_response)

    def _handle_read(
        self,
        path: str,
        response: Dict[str, Any],
        http_response: HttpResponse,
        include_timestamp: Optional[bool] = False,
        include_aggregates: Optional[bool] = False,
    ) -> Optional[SerializedTagWithAggregates]:
        if response is None:
            return None

        agg = response.get("aggregates") or {}  # type: Dict[str, Any]
        current = response.get("current", response)  # type: Dict[str, Any]
        if current is None:
            return None

        if "type" in response:
            val = response
        else:
            val = current["value"]
        if val is None:
            return None

        assert "value" in val

        timestamp = None  # type: Optional[datetime.datetime]
        if include_timestamp:
            if current.get("timestamp") is None:
                raise self.invalid_response(http_response)

            timestamp = TimestampUtilities.str_to_datetime(current["timestamp"])

        return SerializedTagWithAggregates(
            path,
            tbase.DataType.from_api_name(val["type"]),
            val["value"],
            timestamp,
            agg.get("count"),
            agg.get("min"),
            agg.get("max"),
            float(agg["avg"]) if agg.get("avg") is not None else None,
        )

    @classmethod
    def invalid_response(cls, response: HttpResponse) -> core.ApiException:
        request = response.request
        return core.ApiException(
            "Invalid response from {} {}".format(request.method, request.url)
        )



# -*- coding: utf-8 -*-

"""Implementation of AsyncTagQueryResultCollection."""

import abc
from typing import List, Optional

from nisystemlink.clients import core, tag as tbase


class AsyncTagQueryResultCollection(abc.ABC):
    """Represents a paginated list of tags returned by an asynchronous query."""

    def __init__(
        self, first_page: List[tbase.TagData], total_count: int, skip: int
    ) -> None:
        """Initialize an instance with the first page of query results.

        Args:
            first_page: The first page of results, or None if there are no results.
            total_count: The total number of results in the query.
            skip: The skip used for the first page of results.
        """
        self._total_count = total_count
        self._current_page = None  # type: Optional[List[tbase.TagData]]
        if first_page:
            if skip >= total_count:
                raise core.ApiException(
                    "skip is >= totalCount, but the tag list isn't empty"
                )
            self._current_page = first_page
        else:
            pass  # leave it as None, even if passed in as []
        self._skip = skip
        self._current_skip = skip

    @property
    def current_page(self) -> Optional[List[tbase.TagData]]:  # noqa: D401
        """The current page of tag results that were last retrieved from the server, or
        None if there are no more results.

        Use :meth:`move_next_page_async()` to get the next page of results.
        """
        return self._current_page

    @property
    def total_count(self) -> int:  # noqa: D401
        """The total number of tags matched by the query at the time the query was made."""
        return self._total_count

    async def move_next_page_async(self) -> Optional[List[tbase.TagData]]:
        """Asynchronously retrieve the next page of query results from the server,
        returning them and updating :attr:`current_page`.

        Does nothing if the last page has already been retrieved. Use
        :meth:`reset_async()` to start again from the first page.

        Returns:
            A task representing the asynchronous operation. On success, contains the
            next page of results, or None if there are no more results.

        Raises:
            ApiException: if the API call fails.
        """
        if self._current_page is None:
            return None

        new_skip = self._current_skip + len(self._current_page)
        if new_skip < self.total_count:
            self._current_page = await self._query_page_async(new_skip)
        else:
            self._current_page = None

        self._current_skip = new_skip
        return self._current_page

    async def reset_async(self) -> List[tbase.TagData]:
        """Asynchronously query the server for a fresh set of results, returning the
        first page and updating :attr:`current_page` and :attr:`total_count`.

        Returns:
            A task representing the asynchronous operation. On success, contains the
            first page of results, or None if there are no results.

        Raises:
            ApiException: if the API call fails.
        """
        self._current_skip = self._skip
        self._current_page = await self._query_page_async(self._current_skip)
        return self._current_page

    @abc.abstractmethod
    async def _query_page_async(self, skip: int) -> List[tbase.TagData]:
        """Asynchronously query for a single page of results and updates :attr:`total_count`.

        Args:
            skip: The skip to use in the query.

        Returns:
            A task representing the asynchronous operation. On completion, the page of
            results, or None if there are no more results.

        Raises:
            ApiException: if the API call fails.
        """
        ...



# -*- coding: utf-8 -*-

"""Implementation of ClasspropertySupport."""

import abc
import typing
from typing import Any, Callable, Dict, Tuple


class ClasspropertySupport(abc.ABCMeta):
    """Meta class for classes that cannot be subclassed."""

    class _ClassProperty:
        def __init__(self, func: Callable[[Any], Any]):
            self.func = func

    def __new__(
        meta, name: str, bases: Tuple[type, ...], dct: Dict[str, Any]
    ) -> "ClasspropertySupport":
        class ClasspropertySupport_(meta):  # type: ignore
            pass

        for k, v in dct.items():
            # Promote all instance properties to be class properties
            if isinstance(v, meta._ClassProperty):
                setattr(ClasspropertySupport_, k, property(v.func))

        return typing.cast(
            ClasspropertySupport,
            super().__new__(ClasspropertySupport_, name, bases, dct),
        )

    @classmethod
    def classproperty(cls, f: Callable[[Any], Any]) -> property:
        """Make a classproperty."""
        # Cast to a property for the type checker, as we'll convert it to a property
        # later, in the __new__ method
        return typing.cast(property, cls._ClassProperty(f))



# -*- coding: utf-8 -*-

"""Implementation of ITagReader."""

import abc
import datetime
import typing
from typing import Any, Callable, Optional

from nisystemlink.clients import core, tag as tbase
from nisystemlink.clients.core._internal._timestamp_utilities import TimestampUtilities
from nisystemlink.clients.tag._core._serialized_tag_with_aggregates import (
    SerializedTagWithAggregates,
)
from typing_extensions import Literal


_DESERIALIZERS = {
    tbase.DataType.BOOLEAN: {"True": True, "False": False}.get,
    tbase.DataType.DATE_TIME: TimestampUtilities.str_to_datetime,
    tbase.DataType.DOUBLE: float,
    tbase.DataType.INT32: int,
    tbase.DataType.UINT64: int,
    tbase.DataType.STRING: str,
}


class _ITagReaderOverloads(abc.ABC):
    """Contains the overloaded methods of ITagReader.

    These overloads exist so that ``mypy`` can validate proper data types are used when
    reading tag values, as long as a literal DataType enum value is given when calling
    :meth:`get_tag_reader`. But they are hidden away in a base class so that they aren't
    documented by Sphinx, because Sphinx doesn't properly recognize the type annotations
    of overloads. See also: https://github.com/sphinx-doc/sphinx/issues/7901
    """

    @typing.overload
    def get_tag_reader(
        self, path: str, data_type: Literal[tbase.DataType.BOOLEAN]
    ) -> "tbase.TagValueReader[bool]":
        pass

    @typing.overload
    def get_tag_reader(
        self, path: str, data_type: Literal[tbase.DataType.DATE_TIME]
    ) -> "tbase.TagValueReader[datetime.datetime]":
        pass

    @typing.overload
    def get_tag_reader(
        self, path: str, data_type: Literal[tbase.DataType.DOUBLE]
    ) -> "tbase.TagValueReader[float]":
        pass

    @typing.overload
    def get_tag_reader(
        self, path: str, data_type: Literal[tbase.DataType.INT32]
    ) -> "tbase.TagValueReader[int]":
        pass

    @typing.overload
    def get_tag_reader(
        self, path: str, data_type: Literal[tbase.DataType.UINT64]
    ) -> "tbase.TagValueReader[int]":
        pass

    @typing.overload
    def get_tag_reader(
        self, path: str, data_type: Literal[tbase.DataType.STRING]
    ) -> "tbase.TagValueReader[str]":
        pass

    @typing.overload
    def get_tag_reader(
        self, path: str, data_type: tbase.DataType
    ) -> "tbase.TagValueReader[Any]":
        pass

    def get_tag_reader(
        self, path: str, data_type: tbase.DataType
    ) -> "tbase.TagValueReader":
        """Get a :class:`TagValueReader` for this path.

        Args:
            path: The path of the tag to read.
            data_type: The data type of the tag to read.
        """
        return self._get_tag_reader(path, data_type)

    @abc.abstractmethod
    def _get_tag_reader(
        self, path: str, data_type: tbase.DataType
    ) -> "tbase.TagValueReader":
        """Get a :class:`TagValueReader` for this path.

        Args:
            path: The path of the tag to read.
            data_type: The data type of the value to read.
        """
        ...


class ITagReader(_ITagReaderOverloads):
    """Provides an interface for reading the current and aggregate values of a single SystemLink tag."""

    def read(
        self,
        path: str,
        *,
        include_timestamp: bool = False,
        include_aggregates: bool = False
    ) -> Optional[tbase.TagWithAggregates]:
        """Retrieve the current value of the tag with the given ``path`` from the server.

        Optionally retrieves the aggregate values as well. The tag must exist.

        Args:
            path: The path of the tag to read.
            include_timestamp: True to include the timestamp associated with the value
                in the result.
            include_aggregates: True to include the tag's aggregate values in the result
                if the tag is set to :attr:`TagData.collect_aggregates`.

        Returns:
            The value, and the timestamp and/or aggregate values if requested, or None
            if the tag exists but doesn't have a value.

        Raises:
            ValueError: if ``path`` is empty or invalid.
            ValueError: if ``path`` is None.
            ReferenceError: if the reader has been closed.
            ApiException: if the API call fails.
        """
        data = self._read(path, include_timestamp, include_aggregates)
        if data is None or data.value is None:
            return None

        value = self._deserialize_value(data.value, data.data_type)
        if value is None:
            # TODO: Error information
            raise core.ApiException()

        return tbase.TagWithAggregates(
            data.path,
            data.data_type,
            value,
            data.timestamp,
            data.count,
            self._deserialize_value(data.min, data.data_type),
            self._deserialize_value(data.max, data.data_type),
            data.mean,
        )

    async def read_async(
        self,
        path: str,
        *,
        include_timestamp: bool = False,
        include_aggregates: bool = False
    ) -> Optional[tbase.TagWithAggregates]:
        """Asynchronously retrieve the current value of the tag with the given ``path`` from the server.

        Optionally retrieves the aggregate values as well. The tag must exist.

        Args:
            path: The path of the tag to read.
            include_timestamp: True to include the timestamp associated with the value
                in the result.
            include_aggregates: True to include the tag's aggregate values in the result
                if the tag is set to :attr:`TagData.collect_aggregates`.

        Returns:
            The value, and the timestamp and/or aggregate values if requested, or None
            if the tag exists but doesn't have a value.

        Raises:
            ValueError: if ``path`` is empty or invalid.
            ValueError: if ``path`` is None.
            ReferenceError: if the reader has been closed.
            ApiException: if the API call fails.
        """
        data = await self._read_async(path, include_timestamp, include_aggregates)
        if data is None or data.value is None:
            return None

        value = self._deserialize_value(data.value, data.data_type)
        if value is None:
            # TODO: Error information
            raise core.ApiException()

        return tbase.TagWithAggregates(
            data.path,
            data.data_type,
            value,
            data.timestamp,
            data.count,
            self._deserialize_value(data.min, data.data_type),
            self._deserialize_value(data.max, data.data_type),
            data.mean,
        )

    @abc.abstractmethod
    def _read(
        self, path: str, include_timestamp: bool, include_aggregates: bool
    ) -> Optional[SerializedTagWithAggregates]:
        """Retrieve the current value of the tag with the given ``path`` from the server.

        Optionally retrieves the aggregate values as well. The tag must exist. Clients
        do not typically call this method directly. Use a :class:`.TagValueReader`
        instead.

        Args:
            path: The path of the tag to read.
            include_timestamp: True to include the timestamp associated with the value
                in the result.
            include_aggregates: True to include the tag's aggregate values in the result
                if the tag is set to :attr:`TagData.collect_aggregates`.

        Returns:
            The value serialized as a string, or None if the tag exists but doesn't have
            a value.

        Raises:
            ValueError: if ``path`` is empty or invalid.
            ValueError: if ``path`` is None.
            ReferenceError: if the reader has been closed.
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    async def _read_async(
        self, path: str, include_timestamp: bool, include_aggregates: bool
    ) -> Optional[SerializedTagWithAggregates]:
        """Asynchronously retrieve the current value of the tag with the given ``path`` from the server.

        Optionally retrieves the aggregate values as well. The tag must exist. Clients
        do not typically call this method directly. Use a :class:`.TagValueReader`
        instead.

        Args:
            path: The path of the tag to read.
            include_timestamp: True to include the timestamp associated with the value
                in the result.
            include_aggregates: True to include the tag's aggregate values in the result
                if the tag is set to :attr:`TagData.collect_aggregates`.

        Returns:
            The value serialized as a string, or None if the tag exists but doesn't have
            a value.

        Raises:
            ValueError: if ``path`` is empty or invalid.
            ValueError: if ``path`` is None.
            ReferenceError: if the reader has been closed.
            ApiException: if the API call fails.
        """
        ...

    @classmethod
    def _deserialize_value(cls, value: Optional[str], data_type: tbase.DataType) -> Any:
        if value is None:
            return None
        try:
            deserializer = typing.cast(Callable[[str], Any], _DESERIALIZERS[data_type])
        except KeyError:
            raise ValueError("data_type is unknown")
        try:
            return deserializer(value)
        except ValueError:
            return None

    def _get_tag_reader(
        self, path: str, data_type: tbase.DataType
    ) -> "tbase.TagValueReader":
        """Get a :class:`TagValueReader` for this path.

        Args:
            path: The path of the tag to read.
            data_type: The data type of the value to read.
        """
        return tbase.TagValueReader(self, tbase.TagData(path, data_type))



from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._data_frame import DataFrame


class TableRows(JsonModel):
    """Contains the result of a query for rows of decimated data."""

    frame: DataFrame
    """The data frame containing the rows of data."""



# -*- coding: utf-8 -*-

"""Implementation of TagDataUpdate."""

from typing import Any, Dict, Iterable, Optional, Tuple

from nisystemlink.clients import tag as tbase
from typing_extensions import final


@final
class TagDataUpdate:
    """Contains information for updating parts of a tag's metadata on the server when
    used with the :meth:`TagManager.update()` method.

    The update can add keywords, add new properties, change the value of existing
    properties, modify the collect aggregates setting, and modify retention settings. To
    remove a keyword or property, pass the entire :class:`TagData` to the
    :meth:`TagManager.update()` method instead.
    """

    def __init_subclass__(cls) -> None:
        raise TypeError("type 'TagDataUpdate' is not an acceptable base type")

    def __init__(
        self,
        path: str,
        data_type: tbase.DataType,
        keywords: Optional[Iterable[str]] = None,
        properties: Optional[Dict[str, str]] = None,
    ) -> None:
        """Initialize an update of a tag's keywords and/or properties.

        Keywords and properties included in the update that are missing from the tag's
        metadata on the server are added, and properties that exist with a different
        value are replaced. At least one of ``keywords`` or ``properties`` must be
        specified.

        Args:
            path: The path of the tag to update.
            data_type: The data type of the tag to update.
            keywords: The list of keywords that will be added to the tag's metadata on
                the server, or None to not add any keywords.
            properties: The properties that will be added to or replaced in the tag's
                metadata on the server, or None to not modify any properties.

        Raises:
            ValueError: if ``path`` is None.
            ValueError: if both ``keywords`` and ``properties`` are None.
        """
        if path is None:
            raise ValueError("path cannot be None")
        self._path = path
        self._data_type = data_type

        if keywords is None and properties is None:
            raise ValueError(
                "Must specify at least one of keywords or properties to update"
            )
        self._keywords = list(keywords) if keywords is not None else None
        self._properties = None  # type: Optional[Dict[str, str]]
        if properties is not None:
            self._properties = dict(properties)

        self._collect_aggregates = None  # type: Optional[bool]

    @classmethod
    def from_tagdata(
        cls, data: tbase.TagData, fields: tbase.TagUpdateFields
    ) -> "TagDataUpdate":
        """Create an update by taking one or more fields from a :class:`TagData`.

        Args:
            data: The metadata to send to the server in the update.
            fields: One or more fields to include in the update.

        Raises:
            ValueError: if ``data`` is None.
            ValueError: if ``fields`` has no fields or invalid fields.
        """
        if data is None:
            raise ValueError("data cannot be None")

        if fields == 0:
            raise ValueError("Must specify at least one field to update")

        if (tbase.TagUpdateFields.ALL & fields) != fields:
            raise ValueError("Invalid field specified")

        obj = cls.__new__(cls)
        obj._path = data.path
        obj._data_type = data.data_type
        obj._keywords = None
        obj._properties = None
        obj._collect_aggregates = None

        if fields & tbase.TagUpdateFields.KEYWORDS and data.keywords:
            obj._keywords = list(data.keywords)

        if fields & tbase.TagUpdateFields.PROPERTIES and data.properties:
            obj._properties = dict(data.properties)
        if fields & tbase.TagUpdateFields.RETENTION:
            if obj._properties is None:
                obj._properties = {}
            data._copy_retention_properties(obj._properties)

        if fields & tbase.TagUpdateFields.COLLECT_AGGREGATES:
            obj._collect_aggregates = data.collect_aggregates

        return obj

    def to_json_dict(self) -> Dict[str, Any]:
        if self.data_type == tbase.DataType.UNKNOWN:
            raise ValueError("Invalid tag data type")

        data = {
            "path": tbase.TagPathUtilities.validate(self._path),
            "type": self._data_type.api_name,
        }  # type: Dict[str, Any]

        if self._keywords is not None:
            data["keywords"] = self._keywords
        if self._properties is not None:
            data["properties"] = self._properties
        if self._collect_aggregates is not None:
            data["collectAggregates"] = self._collect_aggregates
        return data

    @property
    def collect_aggregates(self) -> Optional[bool]:  # noqa: D401
        """The :attr:`TagData.collect_aggregates` setting to send with the update, or None to not send a value."""
        return self._collect_aggregates

    @property
    def data_type(self) -> tbase.DataType:  # noqa: D401
        """The data type for the tag's values."""
        return self._data_type

    @property
    def keywords(self) -> Optional[Tuple[str, ...]]:  # noqa: D401
        """The list of keywords to send with the update, or None to not send any keywords."""
        return tuple(self._keywords) if self._keywords is not None else None

    @property
    def path(self) -> str:  # noqa: D401
        """The tag's path, which uses a dot-separated hierarchy to uniquely identify the tag on the server."""
        return self._path

    @property
    def properties(self) -> Optional[Dict[str, str]]:  # noqa: D401
        """The properties send with the update, or None to not send any properties."""
        if self._properties:
            return dict(self._properties)
        else:
            return None



from nisystemlink.clients.core._api_info import Operation
from nisystemlink.clients.core._uplink._json_model import JsonModel


class V1Operations(JsonModel):
    """The operations available in the routes provided by the v1 HTTP API."""

    create_specifications: Operation
    """The ability to create new specifications."""

    query_specifications: Operation
    """The ability to query specifications."""

    update_specifications: Operation
    """The ability to update specifications."""

    delete_specifications: Operation
    """The ability to delete specifications."""



# -*- coding: utf-8 -*-

"""Implementation of HttpConfigurationFile."""

from typing import Any, Dict, Optional

from typing_extensions import final


@final
class HttpConfigurationFile:
    """Represents a SystemLink HTTP configuration JSON file."""

    def __init_subclass__(cls) -> None:
        raise TypeError("type 'HttpConfigurationFile' is not an acceptable base type")

    __slots__ = [
        "_id",
        "_display_name",
        "_connection_type",
        "_uri",
        "_api_key",
        "_cert_path",
    ]

    def __init__(self) -> None:
        self._id = None  # type: Optional[str]
        self._display_name = None  # type: Optional[str]
        self._connection_type = None  # type: Optional[str]
        self._uri = None  # type: Optional[str]
        self._api_key = None  # type: Optional[str]
        self._cert_path = None  # type: Optional[str]

    @staticmethod
    def from_json_dict(d: Dict[str, Any]) -> "HttpConfigurationFile":
        """Create from the dictionary format that is stored as JSON in the file."""
        self = HttpConfigurationFile()
        self.id = d.get("Id")
        self.display_name = d.get("DisplayName")
        self.connection_type = d.get("ConnectionType")
        self.uri = d.get("Uri")
        self.api_key = d.get("ApiKey")
        self.cert_path = d.get("CertPath")
        return self

    @property
    def id(self) -> Optional[str]:  # noqa: D401
        """The ID of this configuration."""
        return self._id

    @id.setter
    def id(self, value: Optional[str]) -> None:
        self._id = value

    @property
    def display_name(self) -> Optional[str]:  # noqa: D401
        """A user-friendly display name for this configuration."""
        return self._display_name

    @display_name.setter
    def display_name(self, value: Optional[str]) -> None:
        self._display_name = value

    @property
    def connection_type(self) -> Optional[str]:  # noqa: D401
        """The type of connection to use."""
        return self._connection_type

    @connection_type.setter
    def connection_type(self, value: Optional[str]) -> None:
        self._connection_type = value

    @property
    def uri(self) -> Optional[str]:  # noqa: D401
        """The URI of the SystemLink server."""
        return self._uri

    @uri.setter
    def uri(self, value: Optional[str]) -> None:
        self._uri = value

    @property
    def api_key(self) -> Optional[str]:  # noqa: D401
        """The API key of this client."""
        return self._api_key

    @api_key.setter
    def api_key(self, value: Optional[str]) -> None:
        self._api_key = value

    @property
    def cert_path(self) -> Optional[str]:  # noqa: D401
        """The path to the server's HTTPS certificate, relative to the Skyline Certificates directory."""
        return self._cert_path

    @cert_path.setter
    def cert_path(self, value: Optional[str]) -> None:
        self._cert_path = value



# -*- coding: utf-8 -*-

"""Implementation of TagWithAggregates."""

import datetime
import math
from typing import Generic, Optional, TypeVar, Union

from nisystemlink.clients import core, tag as tbase


_NUMERIC_TYPES = set(
    (tbase.DataType.DOUBLE, tbase.DataType.INT32, tbase.DataType.UINT64)
)

_Any = TypeVar("_Any")


class TagWithAggregates(Generic[_Any]):
    """Represents a generic tag value with optional timestamp and optional aggregate values."""

    def __init__(
        self,
        path: str,
        data_type: tbase.DataType,
        value: _Any,
        timestamp: Optional[datetime.datetime] = None,
        count: Optional[int] = None,
        min: Optional[Union[int, float]] = None,
        max: Optional[Union[int, float]] = None,
        mean: Optional[float] = None,
    ) -> None:
        """Initialize an instance.

        Args:
            path: The path of the tag associated with the value.
            data_type: The data type of the value.
            value: The value.
            timestamp: The timestamp associated with the value.
            count: The number of times the tag has been written, or None if the tag is
                not collecting aggregates.
            min: The minimum value of the tag, or None if the tag is not collecting
                aggregates or the data type of the tag does not track a minimum value.
            max: The maximum value of the tag, or None if the tag is not collecting
                aggregates or the data type of the tag does not track a maximum value.
            mean: The mean value of the tag, or None if the tag is not collecting
                aggregates or the data type of the tag does not track a mean value.

        Raises:
            ApiException: if min, max, or mean is None when it shouldn't be or non-None
                when it should be

        :meta private:
        """
        if data_type in _NUMERIC_TYPES:
            if count is not None:
                if mean is None or min is None or max is None:
                    # TODO: Error information
                    raise core.ApiException()
            elif min is not None or max is not None or mean is not None:
                # TODO: Error information: only valid if count is set
                raise core.ApiException()
        else:
            if (
                min is not None
                or max is not None
                or (mean is not None and not math.isnan(mean))
            ):
                # TODO: Error information: not supported for non-numerics
                raise core.ApiException()

        self._path = path
        self._data_type = data_type
        self._value = value
        self._timestamp = timestamp
        self._count = count
        self._min = min
        self._max = max
        self._mean = mean

    @property
    def data_type(self) -> tbase.DataType:  # noqa: D401
        """The data type of the value."""
        return self._data_type

    @property
    def path(self) -> str:  # noqa: D401
        """The path of the tag associated with the value."""
        return self._path

    @property
    def value(self) -> _Any:  # noqa: D401
        """The value of the tag."""
        return self._value

    @property
    def timestamp(self) -> Optional[datetime.datetime]:  # noqa: D401
        """The timestamp associated with the value, if available."""
        return self._timestamp

    @property
    def count(self) -> Optional[int]:  # noqa: D401
        """The number of times the tag has been written, or None if the tag is not collecting aggregates."""
        return self._count

    @property
    def min(self) -> Optional[Union[int, float]]:  # noqa: D401
        """The minimum value of the tag, or None if the tag is not collecting aggregates
        or the data type of the tag does not track a minimum value.
        """
        return self._min

    @property
    def max(self) -> Optional[Union[int, float]]:  # noqa: D401
        """The maximum value of the tag, or None if the tag is not collecting aggregates
        or the data type of the tag does not track a maximum value.
        """
        return self._max

    @property
    def mean(self) -> Optional[float]:  # noqa: D401
        """The mean value of the tag, or None if the tag is not collecting aggregates or
        the data type of the tag does not track a mean value.
        """
        return self._mean



from typing import List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._column_filter import ColumnFilter


class QueryTableDataBase(JsonModel):
    """Contains the common set of options when querying table data."""

    columns: Optional[List[str]] = None
    """The names of columns to include in the response. The response will
    include the columns in the same order specified in this parameter. All
    columns are included in the order specified at table creation if this
    property is excluded."""

    filters: Optional[List[ColumnFilter]] = None
    """A list of columns to filter by. Only rows whose columns contain values
    matching all of the specified filters are returned. The columns used for
    filtering do not need to be included in the columns list. When reading
    decimated data, the filters are applied before decimation."""



from nisystemlink.clients.core._api_info import Operation
from nisystemlink.clients.core._uplink._json_model import JsonModel


class OperationsV1(JsonModel):
    """The operations available in the routes provided by the v1 HTTP API."""

    create_tables: Operation
    """The ability to create new data tables."""

    delete_tables: Operation
    """The ability to delete tables and all of their data."""

    modify_metadata: Operation
    """The ability to modify metadata for tables."""

    list_tables: Operation
    """The ability to locate and read metadata for tables."""

    read_data: Operation
    """The ability to query and read data from tables."""

    write_data: Operation
    """The ability to append rows of data to tables."""


class ApiInfo(JsonModel):
    """Information about the available API operations."""

    operations: OperationsV1



# -*- coding: utf-8 -*-

"""Implementation of ITagWriter."""

import abc
import datetime
import typing
from typing import Any, Awaitable, Dict, Optional, Tuple, Union

from nisystemlink.clients import tag as tbase
from nisystemlink.clients.core._internal._timestamp_utilities import TimestampUtilities
from typing_extensions import Literal


_VALID_TYPES = {
    tbase.DataType.BOOLEAN: bool,
    tbase.DataType.DATE_TIME: datetime.datetime,
    tbase.DataType.DOUBLE: (float, int),
    tbase.DataType.INT32: int,
    tbase.DataType.UINT64: int,
    tbase.DataType.STRING: str,
}  # type: Dict[tbase.DataType, Union[type, Tuple[type, type]]]


class _ITagWriterOverloads(abc.ABC):
    """Contains the overloaded methods of ITagWriter.

    These overloads exist so that ``mypy`` can validate proper data types are used when
    writing tag values, as long as a literal DataType enum value is given when calling
    :meth:`get_tag_writer`. But they are hidden away in a base class so that they aren't
    documented by Sphinx, because Sphinx doesn't properly recognize the type annotations
    of overloads. See also: https://github.com/sphinx-doc/sphinx/issues/7901
    """

    @typing.overload
    def get_tag_writer(
        self, path: str, data_type: Literal[tbase.DataType.BOOLEAN]
    ) -> "tbase.TagValueWriter[bool]":
        pass

    @typing.overload
    def get_tag_writer(
        self, path: str, data_type: Literal[tbase.DataType.DATE_TIME]
    ) -> "tbase.TagValueWriter[datetime.datetime]":
        pass

    @typing.overload
    def get_tag_writer(
        self, path: str, data_type: Literal[tbase.DataType.DOUBLE]
    ) -> "tbase.TagValueWriter[float]":
        pass

    @typing.overload
    def get_tag_writer(
        self, path: str, data_type: Literal[tbase.DataType.INT32]
    ) -> "tbase.TagValueWriter[int]":
        pass

    @typing.overload
    def get_tag_writer(
        self, path: str, data_type: Literal[tbase.DataType.UINT64]
    ) -> "tbase.TagValueWriter[int]":
        pass

    @typing.overload
    def get_tag_writer(
        self, path: str, data_type: Literal[tbase.DataType.STRING]
    ) -> "tbase.TagValueWriter[str]":
        pass

    @typing.overload
    def get_tag_writer(
        self, path: str, data_type: tbase.DataType
    ) -> "tbase.TagValueWriter[Any]":
        pass

    def get_tag_writer(
        self, path: str, data_type: tbase.DataType
    ) -> "tbase.TagValueWriter":
        """Get a :class:`TagValueWriter` for this path.

        Args:
            path: The path of the tag to write.
            data_type: The data type of the tag to write.
        """
        return self._get_tag_writer(path, data_type)

    @abc.abstractmethod
    def _get_tag_writer(
        self, path: str, data_type: tbase.DataType
    ) -> "tbase.TagValueWriter":
        """Get a :class:`TagValueWriter` for this path.

        Args:
            path: The path of the tag to write.
            data_type: The data type of the tag to write.
        """
        ...


class ITagWriter(_ITagWriterOverloads):
    """Provides an interface for writing the current value of a single SystemLink tag."""

    def write(
        self,
        path: str,
        data_type: tbase.DataType,
        value: Union[bool, int, float, str, datetime.datetime],
        *,
        timestamp: Optional[datetime.datetime] = None
    ) -> None:
        """Write a tag's value.

        Args:
            path: The path of the tag to write.
            data_type: The data type of the value to write.
            value: The tag value to write.
            timestamp: A custom timestamp to associate with the value, or None to have
                the server specify the timestamp.

        Raises:
            ValueError: if ``path`` is empty or invalid.
            ValueError: if ``path`` or ``value`` is None.
            ValueError: if ``data_type`` is invalid.
            ValueError: if ``value`` has the wrong data type.
            ReferenceError: if the writer has been closed.
            ApiException: if the API call fails.
        """
        self._validate_type(value, data_type)

        if data_type == tbase.DataType.DATE_TIME:
            value = TimestampUtilities.datetime_to_str(
                typing.cast(datetime.datetime, value)
            )

        self._write(path, data_type, str(value), timestamp)

    def write_async(
        self,
        path: str,
        data_type: tbase.DataType,
        value: str,
        *,
        timestamp: Optional[datetime.datetime] = None
    ) -> Awaitable[None]:
        """Asynchronously write a tag's value.

        Args:
            path: The path of the tag to write.
            data_type: The data type of the value to write.
            value: The tag value to write.
            timestamp: A custom timestamp to associate with the value, or None to have
                the server specify the timestamp.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ValueError: if ``path`` is empty or invalid.
            ValueError: if ``path`` or ``value`` is None.
            ValueError: if ``data_type`` is invalid.
            ValueError: if ``value`` has the wrong data type.
            ReferenceError: if the writer has been closed.
            ApiException: if the API call fails.
        """
        self._validate_type(value, data_type)

        if data_type == tbase.DataType.DATE_TIME:
            value = TimestampUtilities.datetime_to_str(
                typing.cast(datetime.datetime, value)
            )

        return self._write_async(path, data_type, str(value), timestamp)

    @classmethod
    def _validate_type(
        cls,
        value: Union[bool, int, float, str, datetime.datetime],
        data_type: tbase.DataType,
    ) -> None:
        if data_type == tbase.DataType.UNKNOWN:
            raise ValueError("data_type is UNKNOWN")

        if not isinstance(value, _VALID_TYPES[data_type]):
            raise ValueError(
                "value has wrong python data type ({}) for SystemLink data type {}".format(
                    type(value).__name__, data_type.name
                )
            )

        # python's bool is a subclass of int, so the above check won't catch some cases
        if isinstance(value, bool) and data_type != tbase.DataType.BOOLEAN:
            raise ValueError(
                "value has wrong python data type ({}) for SystemLink data type {}".format(
                    type(value).__name__, data_type.name
                )
            )

        if data_type == tbase.DataType.INT32:
            assert isinstance(value, int)
            if not -(2**31) <= value < 2**31:
                raise ValueError(
                    "value {} is not the valid range of an INT32".format(value)
                )
        elif data_type == tbase.DataType.UINT64:
            assert isinstance(value, int)
            if not 0 <= value < 2**64:
                raise ValueError(
                    "value {} is not the valid range of a UINT64".format(value)
                )

    @abc.abstractmethod
    def _write(
        self,
        path: str,
        data_type: tbase.DataType,
        value: str,
        timestamp: Optional[datetime.datetime] = None,
    ) -> None:
        """Write a tag's value that's been serialized to a string.

        Args:
            path: The path of the tag to write.
            data_type: The data type of the value to write.
            value: The tag value to write, serialized as a string.
            timestamp: A custom timestamp to associate with the value, or None to have
                the server specify the timestamp.

        Raises:
            ValueError: if ``path`` is empty or invalid.
            ValueError: if ``path`` or ``value`` is None.
            ValueError: if ``data_type`` is invalid.
            ReferenceError: if the writer has been closed.
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    async def _write_async(
        self,
        path: str,
        data_type: tbase.DataType,
        value: str,
        timestamp: Optional[datetime.datetime] = None,
    ) -> None:
        """Asynchronously write a tag's value that's been serialized to a string.

        Args:
            path: The path of the tag to write.
            data_type: The data type of the value to write.
            value: The tag value to write, serialized as a string.
            timestamp: A custom timestamp to associate with the value, or None to have
                the server specify the timestamp.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ValueError: if ``path`` is empty or invalid.
            ValueError: if ``path`` or ``value`` is None.
            ValueError: if ``data_type`` is invalid.
            ReferenceError: if the writer has been closed.
            ApiException: if the API call fails.
        """
        ...

    def _get_tag_writer(
        self, path: str, data_type: tbase.DataType
    ) -> "tbase.TagValueWriter":
        """Get a :class:`TagValueWriter` for this path.

        Args:
            path: The path of the tag to write.
            data_type: The data type of the tag to write.
        """
        return tbase.TagValueWriter(self, tbase.TagData(path, data_type))



from enum import Enum
from typing import Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class FilterOperation(str, Enum):
    """Represents the different operations that can be used in a filter."""

    Equals = "EQUALS"
    NotEquals = "NOT_EQUALS"
    LessThan = "LESS_THAN"
    LessThanEquals = "LESS_THAN_EQUALS"
    GreaterThan = "GREATER_THAN"
    GreaterThanEquals = "GREATER_THAN_EQUALS"
    Contains = "CONTAINS"
    NotContains = "NOT_CONTAINS"


class ColumnFilter(JsonModel):
    """A filter to apply to the table data."""

    column: str
    """The name of the column to use for filtering."""

    operation: FilterOperation
    """How to compare the column's value with the specified value.

    An error is returned if the column's data type does not support the specified operation:
    * String columns only support ``EQUALS``, ``NOT_EQUALS``, ``CONTAINS``, and ``NOT_CONTAINS``.
    * Non-string columns do not support ``CONTAINS`` or ``NOT_CONTAINS``.
    * When ``value`` is ``None``, the operation must be ``EQUALS`` or ``NOT_EQUALS``.
    * When ``value`` is ``NaN`` for a floating-point column, the operation must be ``NOT_EQUALS``.
    """

    value: Optional[str]
    """The comparison value to use for filtering. An error will be returned if
    the value cannot be converted to the column's data type."""



# -*- coding: utf-8 -*-

"""Implementation of HttpBufferedTagWriter."""

import datetime
from collections import OrderedDict
from typing import Any, Dict, Optional

from nisystemlink.clients import tag as tbase
from nisystemlink.clients.core._internal._http_client import HttpClient
from nisystemlink.clients.core._internal._timestamp_utilities import TimestampUtilities
from nisystemlink.clients.tag._core._itime_stamper import ITimeStamper
from nisystemlink.clients.tag._core._manual_reset_timer import ManualResetTimer
from typing_extensions import final


@final
class HttpBufferedTagWriter(tbase.BufferedTagWriter):
    def __init_subclass__(cls) -> None:
        raise TypeError("type 'HttpBufferedTagWriter' is not an acceptable base type")

    def __init__(
        self,
        client: HttpClient,
        stamper: ITimeStamper,
        buffer_size: int,
        flush_timer: ManualResetTimer,
    ) -> None:
        super().__init__(stamper, buffer_size, flush_timer)
        self._api = client.at_uri("/nitag/v2")
        self._buffer = OrderedDict()  # type: OrderedDict[str, Dict[str, Any]]

    def _buffer_value(self, path: str, value: Dict[str, Any]) -> None:
        if path not in self._buffer:
            self._buffer.setdefault(path, {"path": path, "updates": []})
        self._buffer[path]["updates"].append(value)

    def _clear_buffer(self) -> None:
        self._buffer.clear()

    def _copy_buffer(self) -> Dict[str, Dict[str, Any]]:
        updates = self._buffer
        self._buffer = OrderedDict()
        return updates

    def _create_item(
        self,
        path: str,
        data_type: tbase.DataType,
        value: str,
        timestamp: Optional[datetime.datetime] = None,
    ) -> Dict[str, Any]:
        item = {
            "value": {"value": value, "type": data_type.api_name}
        }  # type: Dict[str, Any]
        if timestamp is not None:
            item["timestamp"] = TimestampUtilities.datetime_to_str(timestamp)
        return item

    def _send_writes(self, updates: Dict[str, Dict[str, Any]]) -> None:
        self._api.post("/update-current-values", data=list(updates.values()))

    async def _send_writes_async(self, updates: Dict[str, Any]) -> None:
        await self._api.as_async.post(
            "/update-current-values", data=list(updates.values())
        )



from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._condition import Condition


class SpecificationLimit(JsonModel):
    """A limit for a specification.

    The limit is the value that a measurement should be compared against during analysis to
    determine the health or pass/fail status of that measurement.
    """

    min: Optional[float] = None
    """Minimum limit of the specification.

    All measurements that map to this specification should be > this limit.
    """

    typical: Optional[float] = None
    """Typical value of the specification."""

    max: Optional[float] = None
    """Maximum value of the specification.

    All measurements that map to this specification should be < this limit.
    """


class SpecificationType(Enum):
    """The overall type of the specification."""

    PARAMETRIC = "PARAMETRIC"
    """Parametric specs have limits."""

    FUNCTIONAL = "FUNCTIONAL"
    """Functional specs only have pass/fail status."""


class SpecificationUserManaged(JsonModel):
    product_id: str
    """Id of the product to which the specification will be associated."""

    spec_id: str
    """User provided value using which the specification will be identified.

    This should be unique for a product and workspace combination.
    """

    workspace: Optional[str] = None
    """Id of the workspace to which the specification will be associated.

    Default workspace will be taken if the value is not given.
    """


class SpecificationServerManaged(JsonModel):
    id: str
    """The global Id of the specification."""

    version: int
    """
    Current version of the specification.

    When an update is applied, the version is automatically incremented.
    """


class SpecificationDefinition(SpecificationUserManaged):

    name: Optional[str] = None
    """Name of the specification."""

    category: Optional[str] = None
    """Category of the specification."""

    type: SpecificationType
    """Type of the specification."""

    symbol: Optional[str] = None
    """Short form identifier of the specification."""

    block: Optional[str] = None
    """Block name of the specification.

    Typically a block is one of the subsystems of the overall product being specified.
    """

    limit: Optional[SpecificationLimit] = None
    """The limits for this spec."""

    unit: Optional[str] = None
    """Unit of the specification."""

    conditions: Optional[List[Condition]] = None
    """Conditions associated with the specification."""

    keywords: Optional[List[str]] = None
    """Keywords or phrases associated with the specification."""

    properties: Optional[Dict[str, str]] = None
    """Additional properties associated with the specification."""


class Specification(SpecificationDefinition, SpecificationServerManaged):
    """The complete definition of a specification."""


class SpecificationCreation(JsonModel):
    """When the spec was created and when."""

    created_at: Optional[datetime] = None
    """ISO-8601 formatted timestamp indicating when the specification was created."""

    created_by: Optional[str] = None
    """Id of the user who created the specification."""


class SpecificationUpdated(JsonModel):
    """When the spec was updated and when."""

    updated_at: Optional[datetime] = None
    """ISO-8601 formatted timestamp indicating when the specification was last updated."""

    updated_by: Optional[str] = None
    """Id of the user who last updated the specification."""


class SpecificationWithHistory(
    Specification, SpecificationCreation, SpecificationUpdated
):
    """A full specification with update and create history."""



# -*- coding: utf-8 -*-

"""Implementation of ApiException."""

import typing
from typing import Optional

from nisystemlink.clients import core


class ApiException(Exception):
    """Represents errors that occur when calling SystemLink APIs."""

    def __init__(
        self,
        message: Optional[str] = None,
        error: Optional[core.ApiError] = None,
        http_status_code: Optional[int] = None,
        inner: Optional[Exception] = None,
    ) -> None:
        """Initialize an exception.

        Args:
            message: The message describing the error.
            error: The error returned by the API.
            http_status_code: The HTTP status code, if this exception was the result of
                an HTTP error.
            inner: The inner exception that caused the error.
        """
        self._message = message
        self._error = error
        self._http_status_code = http_status_code
        self._inner = inner

    @property
    def message(self) -> Optional[str]:  # noqa:D401
        """The error message."""
        return self._message

    @property
    def error(self) -> Optional[core.ApiError]:  # noqa: D401
        """The error information returned by the SystemLink API, or None if the API did
        not return one or the error occurred trying to call the API.
        """
        if self._error is None:
            return None
        else:
            return self._error.copy()

    @property
    def http_status_code(self) -> Optional[int]:  # noqa: D401
        """The HTTP status code, if this exception was the result of an HTTP error."""
        return self._http_status_code

    @property
    def inner_exception(self) -> Optional[Exception]:  # noqa: D401
        """The exception that caused this failure, if any."""
        return self._inner

    def __str__(self) -> str:
        txt = self._message or "API exception occurred"
        if self._error:
            txt += "\n\n" + str(self._error)
        return txt

    def __eq__(self, other: object) -> bool:
        other_ = typing.cast(ApiException, other)
        return all(
            (
                isinstance(other, ApiException),
                self._message == other_._message,
                self._error == other_._error,
                self._http_status_code == other_._http_status_code,
                self._inner == other_._inner,
            )
        )

    def __hash__(self) -> int:
        return hash((self._message, self._error, self._inner))



"""Implementation of TestMonitor Client"""

from typing import Optional

from nisystemlink.clients import core
from nisystemlink.clients.core._uplink._base_client import BaseClient
from nisystemlink.clients.core._uplink._methods import get

from . import models


class TestMonitorClient(BaseClient):
    def __init__(self, configuration: Optional[core.HttpConfiguration]):
        if configuration is None:
            configuration = core.JupyterHttpConfiguration()
        super().__init__(configuration, base_path="/nitestmonitor/v2/")

    @get("")
    def api_info(self) -> models.ApiInfo:
        """Get information about the available API operations.

        Returns:
            Information about available API operations.

        Raises:
            ApiException: if unable to communicate with the `nitestmonitor` service.
        """
        ...



# -*- coding: utf-8 -*-

"""Implementation of TemporaryTagSelection."""

from types import TracebackType
from typing import Awaitable, List, Optional, Type

from nisystemlink.clients import core, tag as tbase
from nisystemlink.clients.core._internal._http_client import HttpClient
from typing_extensions import final


@final
class TemporaryTagSelection:
    """Manages the lifetime of short-lived tag selections that are used within a single API call.

    Closing the instance deletes the selection.
    """

    _TEMPORARY_SELECTION_TIMEOUT_SECONDS = 30
    """Inactivity timeout for short-lived selections created within a single API call."""

    __MAGIC = object()

    def __init_subclass__(cls) -> None:
        raise TypeError("type 'TemporaryTagSelection' is not an acceptable base type")

    @classmethod
    def create(cls, client: HttpClient, paths: List[str]) -> "TemporaryTagSelection":
        """Create a temporary tag selection containing the given paths.

        Args:
            client: The HTTP client object for communicating with the server.
            paths: The tag search paths to include in the selection.
        """
        selection = TemporaryTagSelection(cls.__MAGIC, client)
        selection._create(paths)
        return selection

    @classmethod
    async def create_async(
        cls, client: HttpClient, paths: List[str]
    ) -> "TemporaryTagSelection":
        """Asynchronously create a temporary tag selection containing the given paths.

        Args:
            client: The HTTP client object for communicating with the server.
            paths: The tag search paths to include in the selection.

        Returns:
            A task representing the asynchronous operation. On completion, contains the
            created selection.
        """
        selection = TemporaryTagSelection(cls.__MAGIC, client)
        await selection._create_async(paths)
        return selection

    def __init__(self, magic: object, client: HttpClient) -> None:
        assert (
            magic == self.__MAGIC
        ), "Do not construct a TemporaryTagSelection directly. Use create() instead."
        self._api = client.at_uri("/nitag/v2/selections")
        self._id = None  # type: Optional[str]

    @property
    def id(self) -> Optional[str]:  # noqa: D401
        """The ID of the selection."""
        return self._id

    def close(self) -> None:
        """Close the selection."""
        if self._id is None:
            return

        try:
            self._api.delete("/{id}", params={"id": self._id})
        except core.ApiException:
            pass

        self._id = None

    async def close_async(self) -> None:
        """Asynchronously close the selection.

        Returns:
            A task representing the asynchronous operation.
        """
        if self._id is None:
            return

        try:
            await self._api.as_async.delete("/{id}", params={"id": self._id})
        except core.ApiException:
            pass

        self._id = None

    def __enter__(self) -> "TemporaryTagSelection":
        return self

    async def __aenter__(self) -> "TemporaryTagSelection":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()

    def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Awaitable[None]:
        return self.close_async()

    def __del__(self) -> None:
        self.close()

    def _create(self, paths: List[str]) -> None:
        selection, http_response = self._api.post(
            "",
            data={
                "searchPaths": paths,
                "inactivityTimeout": self._TEMPORARY_SELECTION_TIMEOUT_SECONDS,
            },
        )

        if selection is None or selection.get("id") is None:
            raise tbase.TagManager.invalid_response(http_response)

        self._id = selection["id"]

    async def _create_async(self, paths: List[str]) -> None:
        selection, http_response = await self._api.as_async.post(
            "",
            data={
                "searchPaths": paths,
                "inactivityTimeout": self._TEMPORARY_SELECTION_TIMEOUT_SECONDS,
            },
        )

        if selection is None or selection.get("id") is None:
            raise tbase.TagManager.invalid_response(http_response)

        self._id = selection["id"]



from ._append_table_data_request import AppendTableDataRequest
from ._api_info import ApiInfo, Operation, OperationsV1
from ._create_table_request import CreateTableRequest
from ._column import Column
from ._column_filter import FilterOperation, ColumnFilter
from ._column_order_by import ColumnOrderBy
from ._column_type import ColumnType
from ._data_frame import DataFrame
from ._data_type import DataType
from ._delete_tables_partial_success import DeleteTablesPartialSuccess
from ._export_table_data_request import ExportTableDataRequest, ExportFormat
from ._modify_tables_partial_success import ModifyTablesPartialSuccess
from ._modify_table_request import ColumnMetadataPatch, ModifyTableRequest
from ._modify_tables_request import ModifyTablesRequest, TableMetadataModification
from ._order_by import OrderBy
from ._paged_tables import PagedTables
from ._paged_table_rows import PagedTableRows
from ._query_decimated_data_request import (
    DecimationMethod,
    DecimationOptions,
    QueryDecimatedDataRequest,
)
from ._query_table_data_request import QueryTableDataRequest
from ._query_tables_request import QueryTablesRequest
from ._table_metadata import TableMetadata
from ._table_rows import TableRows

# Alias to provide backwards compatibility for misnamed class, fixed in 1.0.2
TableMetdataModification = TableMetadataModification

# flake8: noqa



from pydantic import BaseModel, Extra


def _camelcase(s: str) -> str:
    """Convert a snake case string to camelCase."""
    parts = iter(s.split("_"))
    return next(parts) + "".join(i.title() for i in parts)


class JsonModel(BaseModel):
    """Base class for models that are serialized to and from JSON."""

    class Config:
        alias_generator = _camelcase
        allow_population_by_field_name = True
        extra = Extra.ignore



# -*- coding: utf-8 -*-

"""Implementation of TagSelection."""

import abc
import asyncio
import datetime
from types import TracebackType
from typing import Awaitable, Dict, List, Optional, Sequence, Tuple, Type, Union

from nisystemlink.clients import tag as tbase
from nisystemlink.clients.tag._core._serialized_tag_with_aggregates import (
    SerializedTagWithAggregates,
)


class TagSelection(tbase.ITagReader):
    """Represents a set of tags that can be read, written, or deleted together.

    Tags may be specified using glob-style wildcards to include multiple tags with a
    common path. Call :meth:`close()` to free resources. Tag reads are buffered, and the
    latest values are only retrieved on first read or by calling
    :meth:`refresh_values()`. Reads for tags that aren't in the selection return None,
    even if the tag exists on the server.

    Note that :class:`TagSelection` objects support using the ``with`` statement (or the
    ``async with`` statement), to :meth:`close()` the selection automatically on exit.
    """

    def __init__(
        self, tags: Sequence[tbase.TagData], paths: Optional[Sequence[str]] = None
    ) -> None:
        """Initialize a selection using queried or existing tag data.

        Args:
            tags: The tags to store in the selection.
            paths: The paths used to query the tags. If left as None, the paths will be
                extracted from ``tags``.

        Raises:
            ValueError: if ``tags`` is None or empty.
            ValueError: if ``tags`` or ``paths`` contains duplicate tags.
        """
        self._closed = False

        if tags is None:
            raise ValueError("tags cannot be None")
        if any(t is None for t in tags):
            raise ValueError("tags cannot contain None")
        if any(not t.path for t in tags):
            raise ValueError("tags cannot contain a tag with an empty path")

        self._metadata = {m.path: m for m in tags}
        if len(self._metadata) != len(tags):
            raise ValueError("tags contains duplicate paths")

        if paths is None:
            self._paths = set(self._metadata.keys())
        else:
            if any(p is None for p in paths):
                raise ValueError("paths cannot contain None")
            self._paths = set(paths)
            if len(self._paths) != len(paths):
                raise ValueError("paths contains duplicates")

        self._readers = {
            r.path: r
            for r in (self._create_value_reader(m) for m in self._metadata.values())
            if r is not None
        }

        self._values = None  # type: Optional[Dict[str, SerializedTagWithAggregates]]

    @property
    def paths(self) -> Tuple[str, ...]:  # noqa: D401
        """The paths of all tags in the selection, including those that do not exist on the server.

        When using wildcards, the original paths with the wildcards are included in the
        list, not the paths matched by the wildcards.
        """
        return tuple(self._paths)

    @property
    def metadata(self) -> Dict[str, tbase.TagData]:  # noqa: D401
        """The most recently retrieved metadata for tags in the selection, indexed by :attr:`TagData.path`.

        Tags in the selection that do not exist on the server will not appear in the
        collection. Call :meth:`refresh_metadata()` to update the data contained in the
        collection.
        """
        return dict(self._metadata)

    @property
    def values(self) -> Dict[str, tbase.TagValueReader]:  # noqa: D401
        """A :class:`.TagValueReader` for reading the most recently retrieved value for
        tags in the selection, indexed by :attr:`.TagValueReader.path`.

        Tags in the selection that do not exist on the server will not appear in the
        collection. Readers for tags without values on the server will return None when
        read. Call :meth:`refresh_values()` to update the values returned by the readers
        in the dictionary.
        """
        return dict(self._readers)

    @abc.abstractmethod
    def _close_internal(self) -> None:
        """Clean up server resources associated with the selection."""
        ...

    @abc.abstractmethod
    async def _close_internal_async(self) -> None:
        """Asynchronously clean up server resources associated with the selection.

        Returns:
            A task representing the asynchronous operation.
        """
        ...

    @abc.abstractmethod
    def _create_subscription_internal(
        self, update_interval: Optional[datetime.timedelta] = None
    ) -> tbase.TagSubscription:
        """Subscribe to receive events when tags in the selection are written to using the specified update interval.

        Args:
            update_interval: How often to receive tag update notifications from the
                server, or None to use the default.

        Returns:
            The created subscription.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    async def _create_subscription_internal_async(
        self, update_interval: Optional[datetime.timedelta] = None
    ) -> tbase.TagSubscription:
        """Asynchronously subscribe to receive events when tags in the selection are
        written to using the specified update interval.

        Args:
            update_interval: How often to receive tag update notifications from the
                server, or None to use the default.

        Returns:
            A task representing the asynchronous operation. On success, contains the
            created subscription.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    def _delete_tags_from_server_internal(self) -> None:
        """Delete all tags in the selection from the server.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    async def _delete_tags_from_server_internal_async(self) -> None:
        """Asynchronously delete all tags in the selection from the server.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    def _read_tag_metadata(self) -> List[tbase.TagData]:
        """Retrieve the metadata of all tags in the selection.

        Returns:
            The list of tag metadata.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    def _read_tag_values(self) -> List[Optional[SerializedTagWithAggregates]]:
        """Retrieve the values of all tags in the selection.

        Returns:
            The list of the tag values.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    def _read_tag_metadata_and_values(
        self,
    ) -> Tuple[List[tbase.TagData], List[Optional[SerializedTagWithAggregates]]]:
        """Retrieve the metadata and values of all tags in the selection.

        Returns:
            A tuple with the list of tag metadata and the list of tag values.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    async def _read_tag_metadata_async(self) -> List[tbase.TagData]:
        """Asynchronously retrieve the metadata of all tags in the selection.

        Returns:
            A task representing the asynchronous operation. When complete, contains the
            list of tag metadata.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    async def _read_tag_values_async(
        self,
    ) -> List[Optional[SerializedTagWithAggregates]]:
        """Asynchronously retrieve the values of all tags in the selection.

        Returns:
            A task representing the asynchronous operation. When complete, contains the
            list of the tag values.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    async def _read_tag_metadata_and_values_async(
        self,
    ) -> Tuple[List[tbase.TagData], List[Optional[SerializedTagWithAggregates]]]:
        """Asynchronously retrieve the metadata and values of all tags in the selection.

        Returns:
            A task representing the asynchronous operation. When complete, contains a
            tuple with the list of tag metadata and the list of tag values.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    def _reset_aggregates_internal(self) -> None:
        """Reset the aggregate values of tags in the selection.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    @abc.abstractmethod
    async def _reset_aggregates_internal_async(self) -> None:
        """Asynchronously reset the aggregate values of tags in the selection.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ApiException: if the API call fails.
        """
        ...

    def _on_paths_changed(self) -> None:
        """Note when the contents of :attr:`paths` is modified.

        The default implementation does nothing.
        """
        pass

    def add_tags(self, tags: List[tbase.TagData]) -> None:
        """Add one or more tags to the selection.

        Tags that are already in the selection are ignored. The tags as given are
        immediately available in the :attr:`metadata` collection. Use
        :meth:`refresh_metadata()` to get the latest data for the tags. The tags will be
        available in the :attr:`values` collection but won't have latest a latest value
        until :meth:`refresh_values()` is called.

        Args:
            tags: The tags to add to the selection.

        Raises:
            ValueError: if any of the given tags are None or have an invalid path.
            ValueError: if ``tags`` is None.
            ReferenceError: if the selection has been closed.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        for tag in self._validate_tags(tags):
            self._paths.add(tag.path)

            # Don't overwrite existing.
            if tag.path not in self._metadata:
                self._metadata[tag.path] = tag
                reader = self._create_value_reader(tag)
                if reader is not None:
                    self._readers[tag.path] = reader

        self._on_paths_changed()

    def clear_tags(self) -> None:
        """Remove all tags from the selection.

        Raises:
            ReferenceError: if the selection has been closed.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        self._paths.clear()
        self._metadata.clear()
        self._readers.clear()

        self._on_paths_changed()

    def close(self) -> None:
        """Clean up server resources associated with the selection."""
        if self._closed:
            return

        self._close_internal()
        self._closed = True
        paths = getattr(self, "_paths", None)
        if paths is not None:
            paths.clear()
        metadata = getattr(self, "_metadata", None)
        if metadata is not None:
            metadata.clear()
        readers = getattr(self, "_readers", None)
        if readers is not None:
            readers.clear()
        values = getattr(self, "_values", None)
        if values is not None:
            values.clear()

    async def close_async(self) -> None:
        """Asynchronously clean up server resources associated with the selection.

        Returns:
            A task representing the asynchronous operation.
        """
        if self._closed:
            return

        await self._close_internal_async()
        self._closed = True
        self._paths.clear()
        self._metadata.clear()
        self._readers.clear()
        if self._values is not None:
            self._values.clear()

    def create_subscription(
        self, *, update_interval: Optional[datetime.timedelta] = None
    ) -> tbase.TagSubscription:
        """Subscribe to receive events when tags in the selection are written to.

        Updates will be queried from the server using the specified or default update
        interval.

        The subscription will include any tags that are currently included in the
        selection when the subscription is created. That list can be seen via the
        :attr:`metadata` property. The subscription will also attempt to include any
        non-wildcard paths that are in the selection's current list of :attr:`paths`.
        Often, the list of :attr:`paths` directly coincides with the :attr:`metadata`,
        but the former may contain paths that did not exist when the selection's
        metadata was last updated from the server, e.g. via :meth:`refresh()`.

        Closing, adding tags, or removing tags from the selection will not affect
        previously created subscriptions.

        Args:
            update_interval: How often to receive tag updates notifications from the
                server. Default is ``datetime.timedelta(seconds=30)``.

        Returns:
            The created subscription.

        Raises:
            ValueError: if update_interval is negative.
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        if update_interval is not None and update_interval.total_seconds() < 0:
            raise ValueError("update_interval cannot be negative")

        return self._create_subscription_internal(update_interval)

    def create_subscription_async(
        self, *, update_interval: Optional[datetime.timedelta] = None
    ) -> Awaitable[tbase.TagSubscription]:
        """Asynchronously subscribe to receive events when tags in the selection are written to.

        Updates will be queried from the server using the specified or default update
        interval.

        Closing, adding tags, or removing tags from the selection will not affect
        previously created subscriptions.

        Args:
            update_interval: How often to receive tag updates notifications from the
                server. Depending on the :class:`TagManager` implementation in use, this
                may involve polling the server or have a minimum value.

        Returns:
            A task representing the asynchronous operation. On success, contains the
            created subscription.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            f = asyncio.get_event_loop().create_future()
            f.set_exception(ReferenceError("TagSelection"))
            return f

        if update_interval and update_interval.total_seconds() < 0:
            f = asyncio.get_event_loop().create_future()
            f.set_exception(ValueError("update_interval cannot be negative"))
            return f

        return self._create_subscription_internal_async(update_interval)

    def delete_tags_from_server(self) -> None:
        """Delete all tags in the selection from the server.

        The tags are not removed from the selection but are removed from
        :attr:`metadata` and :attr:`values`. If any of the tags are recreated, a call to
        :meth:`refresh_metadata()` will restore them in the collection of tags.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        self._delete_tags_from_server_internal()
        self._metadata.clear()
        self._readers.clear()
        if self._values is not None:
            self._values.clear()

    async def delete_tags_from_server_async(self) -> None:
        """Asynchronously delete all tags in the selection from the server.

        The tags are not removed from the selection but are removed from
        :attr:`metadata` and :attr:`values`. If any of the tags are recreated, a call to
        :meth:`refresh_metadata()` will restore them in the collection of tags.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        await self._delete_tags_from_server_internal_async()
        self._metadata.clear()
        self._readers.clear()
        if self._values is not None:
            self._values.clear()

    def open_tags(self, paths: List[str]) -> None:
        """Add one or more tags to the selection by path.

        Tags that are already in the selection are ignored. The tags will not be
        available in the :attr:`metadata` collection until :meth:`refresh_metadata()` is
        called or the :attr:`values` collection until :meth:`refresh_values()` is
        called.

        Args:
            paths: The tag paths to add to the selection. May include glob-style
                wildcards.

        Raises:
            ValueError: if any of the given paths are None or invalid.
            ValueError: if ``paths`` is None.
            ReferenceError: if the selection has been closed.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        self._paths.update(self._validate_paths(paths))
        self._on_paths_changed()

    def refresh(self) -> None:
        """Refresh both :class:`TagData` and current values for all tags in the
        selection, updating :attr:`metadata` and :attr:`values` accordingly.

        Call :meth:`refresh_metadata()` or :meth:`refresh_values()` to only partially
        refresh.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        (metadata, values) = self._read_tag_metadata_and_values()
        self._update_metadata(metadata)
        self._update_values(values)

    async def refresh_async(self) -> None:
        """Asynchronously refresh both the :class:`TagData` and current values for all
        tags in the selection, updating :attr:`metadata` and :attr:`values` accordingly.

        Call :meth:`refresh_metadata_async()` or :meth:`refresh_values_async()` to only
        partially refresh.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        (metadata, values) = await self._read_tag_metadata_and_values_async()
        self._update_metadata(metadata)
        self._update_values(values)

    def refresh_metadata(self) -> None:
        """Refresh the :class:`TagData` for all tags in the selection, updating :attr:`metadata` accordingly.

        :attr:`values` will also be updated with new and removed tags, but those readers
        will continue to return previously available values until
        :meth:`refresh_values()` is called.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        metadata = self._read_tag_metadata()
        self._update_metadata(metadata)

    async def refresh_metadata_async(self) -> None:
        """Asynchronously refresh the :class:`TagData` for all tags in the selection,
        updating :attr:`metadata` accordingly.

        :attr:`values` will also be updated with new and removed tags, but those readers
        will continue to return previously available values until
        :meth:`refresh_values()` is called.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        metadata = await self._read_tag_metadata_async()
        self._update_metadata(metadata)

    def refresh_values(self) -> None:
        """Refresh the current value of tags in the selection returned by the readers in the :attr:`values` collection.

        Readers will return None for tags that no longer exist on the server, or exist
        but haven't yet been written to. New readers will be added to the collection for
        tags that have values but have not yet been added to :attr:`metadata` by a call
        to :meth:`refresh_metadata()`.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        values = self._read_tag_values()
        self._update_values(values)

    async def refresh_values_async(self) -> None:
        """Asynchronously refresh the current value of tags in the selection returned by
        the readers in the :attr:`values` collection.

        Readers will return None for tags that no longer exist on the server, or exist
        but haven't yet been written to. New readers will be added to the collection for
        tags that have values but have not yet been added to :attr:`metadata` by a call
        to :meth:`refresh_metadata()`.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        values = await self._read_tag_values_async()
        self._update_values(values)

    def remove_tags(self, tags: List[Union[tbase.TagData, str]]) -> None:
        """Remove one or more tags from the selection.

        The tags are not removed from the :attr:`metadata` and :attr:`values`
        collections until the next :meth:`refresh_metadata()`.

        Tags not in the selection are ignored. Tags that were added to the selection
        using wildcard paths can only be removed by including the same wildcard paths.
        Tags matched by multiple wildcard paths remain in the selection until all of the
        paths are removed.

        Args:
            tags: The tags to remove from the selection. Either strings (paths) or
                :class:`TagData` objects can be given. For :class:`TagData` objects,
                only the :attr:`TagData.path` is used.

        Raises:
            ValueError: if ``tags`` is None.
            ValueError: if any of the given tags are None or have an invalid path.
            ReferenceError: if the selection has been closed.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        if tags is None:
            raise ValueError("tags cannot be None")

        paths = [t for t in tags if isinstance(t, str)]
        tags_ = [t for t in tags if not isinstance(t, str)]

        self._paths.difference_update(self._validate_paths(paths))
        self._paths.difference_update(t.path for t in self._validate_tags(tags_))

        self._on_paths_changed()

    def reset_aggregates(self) -> None:
        """Reset tag value aggregates on the server for all tags in the selection.

        Tag historical values are not modified. Has no effect on tags that are not set
        to :attr:`TagData.collect_aggregates`. Use :meth:`refresh_values()` to retrieve
        the new aggregates.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            raise ReferenceError("TagSelection")

        self._reset_aggregates_internal()

    def reset_aggregates_async(self) -> Awaitable[None]:
        """Asynchronously reset tag value aggregates on the server for all tags in the selection.

        Tag historical values are not modified. Has no effect on tags that are not set
        to :attr:`TagData.collect_aggregates`. Use :meth:`refresh_values()` to retrieve
        the new aggregates.

        Returns:
            A task representing the asynchronous operation.

        Raises:
            ReferenceError: if the selection has been closed.
            ApiException: if the API call fails.
        """
        if self._closed:
            f = asyncio.get_event_loop().create_future()
            f.set_exception(ReferenceError("TagSelection"))
            return f

        return self._reset_aggregates_internal_async()

    def __enter__(self) -> "TagSelection":
        return self

    async def __aenter__(self) -> "TagSelection":
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        """Clean up resources associated with the selection."""
        self.close()

    def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> Awaitable[None]:
        """Asynchronously clean up resources associated with the selection."""
        return self.close_async()

    def __del__(self) -> None:
        """Finalize resources associated with the selection."""
        self.close()

    def _read(
        self, path: str, include_timestamp: bool, include_aggregates: bool
    ) -> Optional[Optional[SerializedTagWithAggregates]]:
        if self._closed:
            raise ReferenceError("TagSelection")

        if path is None:
            raise ValueError("path cannot be None")

        if self._values is None:
            self.refresh_values()
            assert self._values is not None

        if path in self._values:
            return self._values[path]
        elif path in self._readers:
            return None
        else:
            raise ValueError("Cannot read a tag that is not in the selection")

    async def _read_async(
        self, path: str, include_timestamp: bool, include_aggregates: bool
    ) -> Optional[Optional[SerializedTagWithAggregates]]:
        if self._closed:
            raise ReferenceError("TagSelection")

        if path is None:
            raise ValueError("path cannot be None")

        if self._values is None:
            await self.refresh_values_async()
            assert self._values is not None

        if path in self._values:
            return self._values[path]
        elif path in self._readers:
            return None
        else:
            raise ValueError("Cannot read a tag that is not in the selection")

    def _create_value_reader(
        self, tag: tbase.TagData
    ) -> Optional[tbase.TagValueReader]:
        if tag is None:
            raise ValueError("tag cannot be None")

        if tag.data_type == tbase.DataType.UNKNOWN:
            return None
        else:
            return tbase.TagValueReader(self, tag)

    def _update_metadata(self, metadata: List[tbase.TagData]) -> None:
        missing_paths = set(self._metadata.keys()).difference(t.path for t in metadata)
        for missing_path in missing_paths:
            del self._metadata[missing_path]
            if missing_path in self._readers:
                del self._readers[missing_path]
            if self._values is not None and missing_path in self._values:
                del self._values[missing_path]

        for tag in metadata:
            old_tag = self._metadata.get(tag.path)
            if old_tag is None or old_tag.data_type != tag.data_type:
                reader = self._create_value_reader(tag)
                if reader is not None:
                    self._readers[tag.path] = reader

            self._metadata[tag.path] = tag

    def _update_values(
        self, values: List[Optional[SerializedTagWithAggregates]]
    ) -> None:
        if self._values is None:
            self._values = {}

        missing_values = set(self._values.keys()).difference(
            v.path for v in values if v is not None
        )
        for missing_path in missing_values:
            del self._values[missing_path]

        for value in values:
            if value is None:
                continue

            self._values[value.path] = value

            tag = self._metadata.get(value.path)
            if tag is None:
                tag = tbase.TagData(value.path, value.data_type)
                self._metadata[value.path] = tag
            elif tag.data_type != value.data_type:
                tag.data_type = value.data_type
            else:
                continue

            reader = self._create_value_reader(tag)
            if reader is not None:
                self._readers[value.path] = reader

    @classmethod
    def _validate_paths(cls, paths: List[str]) -> List[str]:
        """Validate that the given tag paths are valid for queries.

        Args:
            paths: The paths to validate.

        Returns:
            The validated paths.

        Raises:
            ValueError: if any of the given paths are None or invalid.
            ValueError: if ``paths`` is None.
        """
        ...
        if paths is None:
            raise ValueError("paths cannot be None")

        for path in paths:
            if path is None:
                raise ValueError("paths cannot contain None")
            tbase.TagPathUtilities.validate_query(path)

        return paths

    @classmethod
    def _validate_tags(cls, tags: List[tbase.TagData]) -> List[tbase.TagData]:
        """Validate that the given tags have valid paths.

        Args:
            tags: The tags to validate.

        Returns:
            The validated tags.

        Raises:
            ValueError: if any of the given tags are None or have an invalid path.
            ValueError: if ``tags`` is None.
        """
        ...
        if tags is None:
            raise ValueError("tags cannot be None")

        for tag in tags:
            if tag is None:
                raise ValueError("tags cannot contain None")
            tag.validate_path()

        return tags



# -*- coding: utf-8 -*-

from typing import Any, Callable, Generator, Iterable, Optional, Union


AnyCallable = Union[Callable[..., Any], Callable[..., None]]


class EventsException(Exception):
    pass


class Events:
    def __init__(self, events: Optional[Iterable[str]] = None) -> None: ...

    def __getattr__(self, name: str) -> Any: ...

    def __repr__(self) -> str: ...

    __str__ = __repr__

    def __len__(self) -> int: ...

    def __iter__(self) -> Generator["_EventSlot", None, None]: ...


class _EventSlot:
    def __init__(self, name: str) -> None: ...

    def __repr__(self) -> str: ...

    def __call__(self, *a: Any, **kw: Any) -> None: ...

    def __iadd__(self, f: AnyCallable) -> "_EventSlot": ...

    def __isub__(self, f: Optional[AnyCallable]) -> "_EventSlot": ...

    def __len__(self) -> int: ...

    def __iter__(self) -> Generator[AnyCallable, None, None]: ...

    def __getitem__(self, key: str) -> AnyCallable: ...
